from pandas import DataFrame


class Tile:
	def __init__(self, data, rows=None, columns=None):
		"""
		:type data: DataFrame or Tile
		:type rows: list[int] or NoneType
		:type columns: list[str] or NoneType
		"""

		self._original_data = data

		if rows is None:
			rows = list(range(len(data)))
		else:
			if max(rows) >= len(data):
				raise ValueError(f'row {max(rows)} does not exist!')
			if min(rows) < 0:
				raise ValueError(f'row {min(rows)} does not exist!')

		if columns is None:
			columns = list(data.columns)
		else:
			missing_columns = [x for x in columns if x not in data.columns]
			if len(missing_columns) > 0:
				raise KeyError(f'columns {missing_columns} are missing!')

		self._rows = rows
		self._columns = columns

	def __eq__(self, other):
		"""
		:type other: Tile
		:rtype: bool
		"""
		return self.rows == other.rows and self.columns == other.columns and self.original_data == other.original_data

	@property
	def original_data(self):
		"""
		:rtype: DataFrame
		"""
		return self._original_data

	@property
	def rows(self):
		"""
		:rtype: list[int]
		"""
		return self._rows

	@property
	def columns(self):
		"""
		:rtype: list[str]
		"""
		return self._columns

	@property
	def num_columns(self):
		"""
		:rtype: int
		"""
		return len(self.columns)

	@property
	def num_rows(self):
		"""
		:rtype: int
		"""
		return len(self.rows)

	@property
	def data(self):
		"""
		:rtype: DataFrame
		"""
		return self.original_data.iloc[self.rows][self.columns]



from .Tile import Tile
from pandas import DataFrame, concat


class Mosaic:
	_STATE_ATTRIBUTES_ = ['_row_groups', '_column_groups', '_original_data']

	def __init__(self, data, row_groups=None, column_groups=None):
		"""
		:type data: DataFrame
		:type row_groups: dict[str, list[int]]
		:type column_groups: dict[str, list[str]]
		"""
		if row_groups is None:
			self._row_groups = {'all_rows': list(range(len(data)))}
		else:
			self._row_groups = {
				key: [rows] if isinstance(rows, (str, int)) else rows
				for key, rows in row_groups.items()
			}

		if column_groups is None:
			self._column_groups = {'all_columns': list(data.columns)}
		else:
			self._column_groups = {
				key: [columns] if isinstance(columns, (str, int)) else columns
				for key, columns in column_groups.items()
			}

		self._original_data = data
		self._tiles = None
		self._create_tiles()

	def _create_tiles(self):
		self._tiles = {
			row_group_name: {
				column_group_name: Tile(data=self.original_data, columns=columns, rows=rows)
				for column_group_name, columns in self._column_groups.items()
			}
			for row_group_name, rows in self._row_groups.items()
		}

	def copy(self):
		"""
		:rtype: Mosaic
		"""
		return self.drop(row_groups=None, column_groups=None)

	def drop(self, row_groups=None, column_groups=None):
		"""
		:type row_groups: str or list[str] or NoneType
		:type column_groups: str or list[str] or NoneType
		:rtype: Mosaic
		"""
		new_row_groups = self._row_groups.copy()
		new_column_groups = self._column_groups.copy()

		if row_groups is not None:
			if isinstance(row_groups, str):
				row_groups = [row_groups]
			for row_group in row_groups:
				del new_row_groups[row_group]
		if column_groups is not None:
			if isinstance(column_groups, str):
				column_groups = [column_groups]
			for column_group in column_groups:
				del new_column_groups[column_group]

		return self.__class__(
			data=self.original_data,
			row_groups=new_row_groups,
			column_groups=new_column_groups
		)

	def get(self, rows, columns):
		"""
		:rtype: Mosaic
		"""
		rows = rows or list(self._row_groups.keys())
		columns = columns or list(self._column_groups.keys())

		row_groups = {x: self._row_groups[x] for x in rows}
		column_groups = {y: self._column_groups[y] for y in columns}
		return self.__class__(data=self.original_data, row_groups=row_groups, column_groups=column_groups)

	def __getstate__(self):
		return {
			name: getattr(self, name)
			for name in self._STATE_ATTRIBUTES_
		}

	def __setstate__(self, state):
		for name, value in state.items():
			setattr(self, name, value)
		self._create_tiles()

	@property
	def original_data(self):
		"""
		:rtype: DataFrame
		"""
		return self._original_data

	@property
	def rows(self):
		"""
		:rtype: list[int]
		"""
		return sum([list(values) for values in self._row_groups.values()], [])

	@property
	def columns(self):
		"""
		:rtype: list[str]
		"""
		return sum([list(values) for values in self._column_groups.values()], [])

	@property
	def data(self):
		"""
		:rtype: DataFrame
		"""
		return self.original_data.iloc[self.rows][self.columns]

	def represent_row(self, row_group_name, max_columns=None, max_rows=None):
		return concat(
			[
				tile.data.iloc[0:min(tile.num_rows, max_rows or tile.num_rows),0:min(tile.num_columns, max_columns or tile.num_columns)]
				for tile in self._tiles[row_group_name].values()
			],
			axis=1, keys=self._tiles[row_group_name].keys()
		)

	def represent_data(self, max_columns=None, max_rows=None):
		"""
		:type max_columns: NoneType or int
		:type max_rows: NoneType or int
		:rtype: DataFrame
		"""

		return concat(
			[
				self.represent_row(row_group_name=row_group_name, max_columns=max_columns, max_rows=max_rows)
				for row_group_name in self._tiles.keys()
			],
			axis=0, keys=self._tiles.keys()
		)

	def display(self, p=None, max_columns=None, max_rows=None):
		representation = self.represent_data(max_columns=max_columns, max_rows=max_rows)

		try:
			from IPython.core.display import display
			if isinstance(representation, dict):
				for key, value in representation.items():
					display(key, value)
			else:
				display(representation)
		except ImportError:
			if p is not None:
				p.pretty(representation)
			else:
				print(representation)

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text('Mosaic')
		else:
			self.display(p=p)

	def __getitem__(self, item):
		if isinstance(item, str) and item in self._row_groups:
			return self.get(rows=[item], columns=None)
		elif isinstance(item, str) and item in self._column_groups:
			return self.get(rows=None, columns=[item])
		elif isinstance(item, (tuple, list)) and len(item) == 2:
			if all([x in self._row_groups for x in item]):
				return self.get(rows=item, columns=None)
			elif all([y in self._column_groups for y in item]):
				return self.get(columns=item, rows=None)
			elif item[0] in self._row_groups and item[1] in self._column_groups:
				return self.get(rows=[item[0]], columns=[item[1]])
			elif item[0] in self._column_groups and item[1] in self._row_groups:
				return self.get(columns=[item[0]], rows=[item[1]])
			else:
				raise KeyError(f'{item}')
		elif isinstance(item, (tuple, list)):
			if all([x in self._row_groups for x in item]):
				return self.get(rows=item, columns=None)
			elif all([y in self._column_groups for y in item]):
				return self.get(columns=item, rows=None)
			else:
				raise KeyError(f'{item}')
		else:
			raise KeyError(f'{item}')



from pensieve import Pensieve
from nightingale.regression import get_regression_model
from ravenclaw.wrangling import bring_to_front
from .Xy import Xy


class XyRegression:
	def __init__(self, xy, interactions=None, groups=None, eliminate=False, regression_type='linear', echo=1):
		"""
		:type xy: Xy
		:type interactions: list[tuple[str]] or NoneType
		"""
		self._pensieve = Pensieve(graph_direction='TB')
		self.pensieve['xy'] = xy
		self.pensieve['input_groups'] = groups
		def get_groups(xy, input_groups):
			if groups:
				return groups
			elif xy.has_id() and xy.has_duplicates():
				return xy.id_columns
			else:
				return None

		self.pensieve['groups'] = get_groups
		self.pensieve['interactions'] = interactions or []
		self.pensieve['elimination'] = eliminate
		self.pensieve['regression_type'] = regression_type

		def get_formula(xy, interactions):
			"""
			:type xy: Xy
			:type interactions: list[tuple[str]] or NoneType
			:rtype: str
			"""
			y = f'{xy.y_column} ~ '
			x = ' + '.join([xy.get_x_column_formula(column=column) for column in xy.x_columns])
			formula = y + x
			for interaction in interactions:
				formula += ' + ' + '*'.join(
					[xy.get_x_column_formula(column=column) for column in interaction if column in xy.x_columns]
				)
			return formula

		self.pensieve['formula'] = get_formula

		def _get_regression_model(formula, xy, regression_type, groups, elimination):
			return get_regression_model(
				formula=formula, data=xy.data, regression_type=regression_type,
				groups=groups, eliminate=elimination, echo=echo
			)
		self.pensieve['regression_model'] = _get_regression_model

		def get_summary(regression_model):
			summary_table = regression_model.summary_table.copy()
			summary_table['model_name'] = regression_model.name

			#summary_table = bring_to_front(data=summary_table, columns=['model_name'])

			if 't' in summary_table.columns:
				summary_table = summary_table.sort_values('t')
			elif 'z' in summary_table.columns:
				summary_table = summary_table.sort_values('z')


			main_columns = ['model_name']


			return bring_to_front(data=summary_table, columns=main_columns).reset_index(drop=True)

		self.pensieve['summary'] = get_summary

	@property
	def pensieve(self):
		"""
		:rtype: Pensieve
		"""
		return self._pensieve

	@property
	def model_name(self):
		"""
		:rtype: str
		"""
		return self.pensieve['model_name']

	@property
	def summary_table(self):
		"""
		:rtype: pandas.DataFrame
		"""
		return self.pensieve['summary']

	@property
	def model(self):
		"""
		:rtype:
		"""
		return self.pensieve['regression_model']

	@property
	def formula(self):
		return self.pensieve['formula']




class TransformedModel:
	def __init__(self, model, transformer):
		self._model = model
		self._transformer = transformer

	def transform(self, data):
		try:
			transformed = self._transformer.predict(data)
		except AttributeError:
			try:
				transformed = self._transformer.pred(data)
			except AttributeError:
				transformed = self._transformer.transform(data)
		return transformed

	def fit(self, data):
		self._transformer.fit(data)
		self._model.fit(self.transform(data=data))

	def predict(self, data):
		transformed = self.transform(data=data)
		try:
			prediction = self._model.predict(transformed)
		except AttributeError:
			prediction = self._model.pred(transformed)
		return prediction



from .Mosaic import Mosaic
from pandas import DataFrame, Series
from ravenclaw import is_numeric
from nightingale.feature_importance import get_model_influence
import warnings

class Xy(Mosaic):
	def __init__(self, data, x_columns, y_column, id_columns=None, row_groups=None):
		"""
		:type data: DataFrame
		:type x_columns: list[str]
		:type y_column: str
		:type id_columns: list[str] or NoneType
		:type row_groups: dict[str, list[int]]
		"""
		if isinstance(x_columns, str):
			x_columns = [x_columns]

		if isinstance(y_column, list) and len(y_column) == 1:
			y_column = y_column[0]

		if isinstance(id_columns, str):
			id_columns = [id_columns]

		if id_columns is None:
			column_groups = {'X': x_columns, 'y': [y_column]}
		else:
			column_groups = {'id': id_columns, 'X': x_columns, 'y': [y_column]}
		super().__init__(data=data, column_groups=column_groups, row_groups=row_groups)

	def drop(self, column_groups=None, row_groups=None):
		"""
		:rtype: Xy
		"""
		new_row_groups = self._row_groups.copy()
		new_column_groups = self._column_groups.copy()
		if row_groups is not None:
			if isinstance(row_groups, str):
				row_groups = [row_groups]
			for row_group in row_groups:
				del new_row_groups[row_group]
		if column_groups is not None:
			if isinstance(column_groups, str):
				column_groups = [column_groups]
			for column_group in column_groups:
				del new_column_groups[column_group]

		return self.__class__(
			data=self.original_data,
			x_columns=new_column_groups['X'].copy(),
			y_column=new_column_groups['y'][0],
			id_columns=new_column_groups.get('id', None),
			row_groups=new_row_groups
		)

	def copy(self):
		"""
		:rtype: Xy
		"""
		return self.drop(column_groups=None, row_groups=None)

	def get(self, rows, columns):
		"""
		:rtype: Mosaic
		"""
		rows = rows or list(self._row_groups.keys())
		row_groups = {x: self._row_groups[x] for x in rows}

		if columns is None:
			return Xy(
				data=self.original_data,
				row_groups=row_groups,
				x_columns=self._column_groups['X'].copy(),
				y_column=self._column_groups['y'][0],
				id_columns=self._column_groups.get('id', None)
			)
		else:
			columns = columns or list(self._column_groups.keys())
			column_groups = {y: self._column_groups[y] for y in columns}
			return Mosaic(data=self.original_data, row_groups=row_groups, column_groups=column_groups)

	@property
	def X(self):
		"""
		:rtype: DataFrame
		"""
		return self['X'].data

	@property
	def y(self):
		"""
		:rtype: Series
		"""
		return self['y'].data.iloc[:, 0]

	@property
	def id(self):
		"""
		:rtype: DataFrame
		"""
		return self['id'].data

	def has_id(self):
		"""
		:rtype: bool
		"""
		return 'id' in self._column_groups

	@property
	def unique_ids(self):
		"""
		:rtype: DataFrame
		"""
		return self['id'].data.drop_duplicates()

	@property
	def num_uniques(self):
		"""
		:rtype: int
		"""
		if self.has_id():
			return len(self.unique_ids)
		else:
			raise RuntimeError('Xy does not have id columns')

	def has_duplicates(self):
		"""
		:rtype: bool
		"""
		return self.num_uniques < len(self.data)

	@property
	def x_columns(self):
		"""
		:rtype: list[str]
		"""
		return self._column_groups['X']

	@property
	def y_column(self):
		"""
		:rtype: str
		"""
		return self._column_groups['y'][0]

	@property
	def id_columns(self):
		"""
		:rtype: list[str]
		"""
		return self._column_groups.get('id', [])

	def get_column_type(self, column):
		"""
		:type column: str
		:rtype: str
		"""
		if is_numeric(self.data[column]):
			return 'numeric'
		else:
			return 'categorical'

	def get_x_column_formula(self, column):
		"""
		:rtype x_column: str
		:rtype: str
		"""
		return f'C({column})' if self.get_column_type(column=column) == 'categorical' else column

	def get_influence(self, model, x_columns=None, num_points=500, num_threads=1, echo=1):
		with warnings.catch_warnings():
			warnings.simplefilter("ignore")
			return get_model_influence(
				model=model, data=self.X, x_columns=x_columns or self.x_columns, num_points=num_points,
				num_threads=num_threads, echo=echo
			)



from .Xy import Xy
from .get_cross_validation_by_group import get_cross_validation_by_group
from .TrainingTestXy import TransformedTrainingTestXy
from .validate_fold import validate_fold

from nightingale.feature_importance import get_feature_importances
from nightingale.feature_importance import get_coefficients
from nightingale.tuning import ModelGrid
from nightingale.evaluation import evaluate_classification, evaluate_regression
from nightingale import Predictor
from ravenclaw.wrangling import bring_to_front
from copy import deepcopy
from pandas import DataFrame
from joblib import Parallel, delayed
import warnings
from chronometry.progress import ProgressBar, iterate
from slytherin.collections import Dictionary


class ValidationXy(TransformedTrainingTestXy):
	_STATE_ATTRIBUTES_ = TransformedTrainingTestXy._STATE_ATTRIBUTES_ + [
		'_echo', '_num_threads', '_num_validation_splits', '_cross_validation', '_folds', '_random_state'
	]

	def __init__(
			self, data, x_columns, y_column, id_columns=None,
			holdout_ratio=0.2, num_validation_splits=5, random_state=None,
			transformer=None, num_threads=-1, echo=1, cross_validation=None
	):
		self._echo = echo
		self._num_threads = num_threads
		self._num_validation_splits = num_validation_splits
		self._random_state = random_state
		progress_bar = ProgressBar(total=num_validation_splits + 1, echo=self._echo)
		progress_amount = 0
		progress_bar.show(amount=progress_amount, text='preparing validation Xy')

		if cross_validation is None:
			cross_validation = get_cross_validation_by_group(
				data=data, id_columns=id_columns, num_splits=num_validation_splits,
				holdout_ratio=holdout_ratio, random_state=random_state
			)
		self._cross_validation = cross_validation

		super().__init__(
			data=data, x_columns=x_columns, y_column=y_column, id_columns=id_columns,
			training_rows=cross_validation['validation'], training_name='validation',
			test_rows=cross_validation['holdout'], test_name='holdout', transformer=deepcopy(transformer)
		)
		progress_amount += 1
		progress_bar.show(amount=progress_amount, text='preparing folds')

		def transform_training_test_xy(fold):
			return TransformedTrainingTestXy(
				data=data, x_columns=x_columns, y_column=y_column, id_columns=id_columns,
				training_rows=fold['training'], test_rows=fold['test'], transformer=deepcopy(transformer)
			)

		if self.num_threads == 1:
			self._folds = [
				transform_training_test_xy(fold=fold)
				for fold in iterate(cross_validation['folds'], text='preparing folds (single-threaded)')
			]
		else:
			processor = Parallel(n_jobs=self.num_threads, backend='threading', require='sharedmem')
			self._folds = processor(
				delayed(transform_training_test_xy)(fold=fold)
				for fold in iterate(cross_validation['folds'], text='preparing folds (multi-threaded)')
			)

	@property
	def num_threads(self):
		"""
		:rtype: int
		"""
		return self._num_threads

	def replace_data(self, new_data, num_threads=None, echo=None):
		"""
		replaces the data with a new_data but keeps the same cross validation
		:type new_data: DataFrame
		:rtype: ValidationXy
		"""
		num_threads = num_threads or self.num_threads
		return self.__class__(
			data=new_data, x_columns=self.x_columns, y_column=self.y_column, id_columns=self.id_columns,
			holdout_ratio=None, num_validation_splits=self._num_validation_splits, random_state=self._random_state,
			transformer=self.transformer, num_threads=num_threads, echo=echo or self._echo,
			cross_validation=self._cross_validation.copy()
		)

	@property
	def validation(self):
		"""
		:rtype: Xy
		"""
		return self['validation']

	@property
	def holdout(self):
		"""
		:rtype: Xy
		"""
		return self['holdout']

	@property
	def folds(self):
		"""
		:rtype: list[TransformedTrainingTestXy]
		"""
		return self._folds

	def represent_data(self, max_columns=None, max_rows=None):
		"""
		:type max_columns: NoneType or int
		:type max_rows: NoneType or int
		:rtype: dict[str, DataFrame]
		"""

		result = {'full_data': super().represent_data(max_columns=max_columns, max_rows=max_rows)}
		for i, fold in enumerate(self.folds):
			result[f'fold_{i + 1}'] = fold.represent_data(max_columns=max_columns, max_rows=max_rows)
		return result

	def validate(
			self, problem_type, evaluation_function=None, model=None, model_name=None, model_grid=None,
			num_threads=None, return_models=False, raise_error=False, main_metric=None,
			best_model_criteria=None,
			measure_influence=True, num_influence_points=400,
			echo=None
	):
		"""
		:param callable evaluation_function:
		:param str problem_type: either 'regression' of 'classification'
		:param model: a regressor or classifier
		:param str or NoneType model_name:
		:param ModelGrid or NoneType model_grid:
		:param dict[str, list or str or int or float] parameter_grid: a dictionary telling how the grid of models should be built
		:param int num_threads: for parallel computing
		:param bool return_models: whether or not trained models should be returned
		:param bool raise_error:
		:param str main_metric: the metric to be used for choosing the best model
		:param str or list[str] or NoneType other_metrics: other metrics to be included in the aggregate
		:param str best_model_criteria: 'highest' means the model with the highest metric value should be chosen
		:param int or bool or ProgressBar echo:
		:rtype: dict[str,]
		"""
		echo = echo or self._echo
		num_threads = num_threads or self._num_threads

		if model is not None:
			try:
				model_class_name = model.__name__
			except AttributeError:
				model_class_name = None
		else:
			model_class_name = None

		if problem_type.lower().startswith('class'):
			model_name = model_name or model_class_name or 'classifier'
			main_metric = main_metric or 'f1_score'
			best_model_criteria = best_model_criteria or 'highest'
			evaluation_function = evaluation_function or evaluate_classification
		else:
			model_name = model_name or model_class_name or 'regressor'
			main_metric = main_metric or 'rmse'
			best_model_criteria = best_model_criteria or 'lowest'
			evaluation_function = evaluation_function or evaluate_regression

		if model is not None and model_grid is None:
			models = {model_name: model}
			parameters_data = None
			parameters_dictionary = None

		elif model is None and model_grid is not None:
			models = model_grid.models
			parameters_data = model_grid.parameter_table
			parameters_dictionary = model_grid.parameters

		else:
			raise ValueError('either a preset model should be given or a model grid')

		model_folds = [
			{'model_name': model_name, 'model': model, 'fold_num': fold_num + 1, 'fold': fold}
			for model_name, model in models.items() for fold_num, fold in enumerate(self.folds)
		]

		shared_memory = {
			'progress_bar': ProgressBar(total=1 + len(models) * len(self.folds), echo=echo),
			'progress_amount': 0
		}

		with warnings.catch_warnings():
			warnings.simplefilter("ignore")
			if num_threads == 1:
				result = [
					validate_fold(model_fold=model_fold, shared_memory=shared_memory, evaluation_function=evaluation_function) for model_fold in model_folds
				]
			else:
				parallel = Parallel(n_jobs=num_threads, backend='threading', require='sharedmem')
				result = parallel(
					delayed(validate_fold)(
						model_fold=model_fold, shared_memory=shared_memory, evaluation_function=evaluation_function
					)
				  	for model_fold in model_folds
				)

		training = []
		test = []
		trained_models = []
		feature_importances_list = []
		coefficients_list = []

		for record in result:
			model = record['model']
			training_evaluation = record['training_evaluation']
			test_evaluation = record['test_evaluation']
			training_evaluation['fold_num'] = record['fold_num']
			test_evaluation['fold_num'] = record['fold_num']
			training_evaluation['model_name'] = record['model_name']
			test_evaluation['model_name'] = record['model_name']
			training.append(training_evaluation)
			test.append(test_evaluation)

			feature_importances = get_feature_importances(
				model=model, columns=self.x_columns, model_name=record['model_name'], fold_num=record['fold_num'],
				raise_error=raise_error
			)
			if feature_importances is not None:
				feature_importances_list.append(feature_importances)

			coefficients = get_coefficients(
				model=model, columns=self.x_columns, model_name=record['model_name'], fold_num=record['fold_num'],
				raise_error=raise_error
			)
			if coefficients is not None:
				coefficients_list.append(coefficients)

			if return_models:
				trained_models.append(model)

		with warnings.catch_warnings():
			warnings.simplefilter("ignore")

			feature_importances = DataFrame.from_records(feature_importances_list)
			coefficients = DataFrame.from_records(coefficients_list)

			training = bring_to_front(data=DataFrame.from_records(training), columns=['model_name', 'fold_num'])
			test = bring_to_front(data=DataFrame.from_records(test), columns=['model_name', 'fold_num'])
			result = Dictionary({
				'training': training,
				'test': test
			})

			if feature_importances.shape[1] > 0:
				result['feature_importances'] = bring_to_front(
					data=feature_importances, columns=['model_name', 'fold_num']
				)
				result['mean_feature_importances'] = result['feature_importances'].drop(
					columns='fold_num'
				).groupby(['model_name']).mean().reset_index()

			if coefficients.shape[1] > 0:
				result['coefficients'] = bring_to_front(
					data=coefficients, columns=['model_name', 'fold_num']
				)
				result['mean_coefficients'] = result['coefficients'].drop(
					columns='fold_num'
				).groupby(['model_name']).mean().reset_index()

			if return_models:
				result['models'] = trained_models

			shared_memory['progress_amount'] += 1
			shared_memory['progress_bar'].show(amount=shared_memory['progress_amount'], text='validation complete.')

			if main_metric is not None and main_metric in result['test'].columns:
				aggregated_training = training.drop(columns='fold_num').groupby('model_name').mean().reset_index()
				aggregated_test = test.drop(columns='fold_num').groupby('model_name').mean().reset_index()
				aggregated_result = aggregated_training.merge(
					right=aggregated_test, on='model_name', how='outer', suffixes=['_training', '_test']
				)
				if parameters_data is not None:
					aggregated_result = parameters_data.merge(right=aggregated_result, on='model_name', how='outer', suffixes=['_parameter', ''])

				aggregated_result = aggregated_result.sort_values(
					f'{main_metric}_test', ascending=best_model_criteria != 'highest'
				)

				result['aggregate'] = aggregated_result
				best_model_name = aggregated_result.head(1)['model_name'].values[0]

				if parameters_dictionary is not None:
					result['best_model_parameters'] = parameters_dictionary[best_model_name]

				best_model = models[best_model_name]
				trained_best_model = deepcopy(best_model)
				trained_best_model.fit(X=self.validation.X, y=self.validation.y)
				best_model_performance = evaluation_function(
					trained_best_model.predict(self.holdout.X), self.holdout.y
				)
				result['best_model'] = best_model
				result['best_model_trained'] = trained_best_model
				result['best_model_performance'] = best_model_performance

				best_model_feature_importances = get_feature_importances(
					model=trained_best_model, columns=self.validation.x_columns, model_name=best_model_name,
					raise_error=raise_error
				)

				best_model_coefficients = get_coefficients(
					model=trained_best_model, columns=self.validation.x_columns, model_name=best_model_name,
					raise_error=raise_error
				)

				if best_model_feature_importances is not None:
					result['best_model_feature_importances'] = best_model_feature_importances

				if best_model_coefficients is not None:
					result['best_model_coefficients'] = best_model_coefficients

				if measure_influence:
					result['influence'] = self.get_influence(
						model=trained_best_model, num_threads=num_threads, num_points=num_influence_points, echo=echo
					)

		model = result['best_model_trained']
		transformer = self._transformer
		predictor = Predictor(model=model, transformer=transformer, model_is_fitted=True, transformer_is_fitted=True, x_columns=self.x_columns)
		result['predictor'] = predictor
		return result



from sklearn.model_selection import train_test_split, KFold

def _get_id_data(data, id_columns):
	new_data = data.copy()
	new_data['_row_'] = range(len(data))

	if id_columns is None:
		new_data['_id_'] = range(len(new_data))
		id_columns = ['_id_']
	ids = new_data[id_columns].drop_duplicates()
	return new_data, id_columns, ids

def get_training_test_split_by_group(data, id_columns, test_ratio, random_state=None):
	new_data, id_columns, ids = _get_id_data(data=data, id_columns=id_columns)

	id_training, id_test = train_test_split(ids, test_size=test_ratio, random_state=random_state)
	training = id_training.merge(new_data, how='left', on=id_columns)
	test = id_test.merge(new_data, how='left', on=id_columns)
	training_indices = training['_row_']
	test_indices = test['_row_']
	return list(training_indices), list(test_indices)

def get_kfold_by_group(data, id_columns, num_splits=5, random_state=None):
	new_data, id_columns, ids = _get_id_data(data=data, id_columns=id_columns)

	kfold = KFold(n_splits=num_splits, shuffle=True, random_state=random_state)
	folds = [
		{
			'training': list(ids.iloc[training_index].merge(right=new_data, on=id_columns, how='left')['_row_']),
			'test': list(ids.iloc[test_index].merge(right=new_data, on=id_columns, how='left')['_row_'])
		}
		for training_index, test_index in kfold.split(ids)
	]
	return folds

def get_cross_validation_by_group(data, id_columns, num_splits=5, holdout_ratio=0.2, random_state=None):
	if holdout_ratio == 0:
		return {
			'validation': list(range(len(data))), 'holdout': [],
			'folds': get_kfold_by_group(data=data, id_columns=id_columns, num_splits=num_splits, random_state=random_state)
		}

	if num_splits < 2:
		training_index, test_index = get_training_test_split_by_group(data=data, id_columns=id_columns, random_state=random_state)
		return {
			'validation': training_index,
			'holdout': test_index,
			'folds': []
		}

	new_data, id_columns, ids = _get_id_data(data=data, id_columns=id_columns)
	validation_index, holdout_index = get_training_test_split_by_group(
		data=new_data, id_columns=id_columns, test_ratio=holdout_ratio
	)

	validation = new_data.iloc[validation_index][id_columns]
	validation_folds = get_kfold_by_group(data=validation, num_splits=num_splits, random_state=random_state, id_columns=id_columns)

	return {
		'validation': validation_index,
		'holdout': holdout_index,
		'folds': [
			{
				'training': list(
					validation.iloc[fold['training']].merge(right=new_data, how='left', on=id_columns)['_row_']
				),
				'test': list(
					validation.iloc[fold['test']].merge(right=new_data, how='left', on=id_columns)['_row_']
				)
			}
			for fold in validation_folds
		]
	}



import warnings
from copy import deepcopy



def validate_fold(model_fold, shared_memory, evaluation_function):
	"""
	:type fold: TransformedTrainingTestXy
	:type evaluation_function: callable
	:type model:
	:rtype: DataFrame
	"""
	shared_memory['progress_bar'].show(amount=shared_memory['progress_amount'], text='validating ...')
	with warnings.catch_warnings():
		warnings.simplefilter('ignore')

		this_model = deepcopy(model_fold['model'])
		fold = model_fold['fold']
		fold_num = model_fold['fold_num']
		model_name = model_fold['model_name']

		this_model.fit(X=fold.training.X, y=fold.training.y)
		training_evaluation = evaluation_function(this_model.predict(fold.training.X), fold.training.y)
		test_evaluation = evaluation_function(this_model.predict(fold.test.X), fold.test.y)
	shared_memory['progress_amount'] += 1
	return {
		'model': this_model,
		'model_name': model_name,
		'fold_num': fold_num,
		'training_evaluation': training_evaluation,
		'test_evaluation': test_evaluation
	}


from .Xy import Xy
from .TransformedModel import TransformedModel


class TrainingTestXy(Xy):
	def __init__(
			self, data, x_columns, y_column, training_rows=None, test_rows=None, id_columns=None,
			training_name='training', test_name='test'
	):
		super().__init__(
			data=data, x_columns=x_columns, y_column=y_column, id_columns=id_columns,
			row_groups={training_name: training_rows, test_name: test_rows}
		)

	@property
	def training(self):
		"""
		:rtype: Xy
		"""
		return self['training']

	@property
	def test(self):
		"""
		:rtype: Xy
		"""
		return self['test']

class TransformedTrainingTestXy(TrainingTestXy):
	def __init__(
			self, data, x_columns, y_column, training_rows=None, test_rows=None, id_columns=None,
			training_name='training', test_name='test', transformer=None
	):
		self._transformer = transformer
		if isinstance(x_columns, str):
			x_columns = [x_columns]

		if isinstance(y_column, list) and len(y_column) == 1:
			y_column = y_column[0]

		if isinstance(id_columns, str):
			id_columns = [id_columns]

		if transformer is None:
			self._original_x_columns = None
			transformed_data = data
			transformed_x_columns = x_columns
			self._original_training_test = None

		else:
			self._original_x_columns = x_columns.copy()
			self._original_training_test = TrainingTestXy(
				data=data, x_columns=x_columns, y_column=y_column, training_rows=training_rows, test_rows=test_rows,
				id_columns=id_columns, training_name=training_name, test_name=test_name

			)
			self._transformer.fit(self._original_training_test[training_name].X)
			transformed_data = self._transformer.transform(data[x_columns])
			transformed_x_columns = list(transformed_data.columns)

			if id_columns is not None:
				for id_column in id_columns:
					transformed_data[id_column] = data[id_column].values
			transformed_data[y_column] = data[y_column].values

		super().__init__(
			data=transformed_data, x_columns=transformed_x_columns, y_column=y_column,
			training_rows=training_rows, test_rows=test_rows, id_columns=id_columns,
			training_name=training_name, test_name=test_name
		)

	@property
	def transformer(self):
		return self._transformer

	def get_influence(self, model, x_columns=None, num_points=500, num_threads=1, echo=1):
		x_columns = x_columns or self._original_x_columns

		if self.transformer is not None:
			transformed_model = TransformedModel(model=model, transformer=self.transformer)
		else:
			transformed_model = model

		return super().get_influence(
			model=transformed_model, x_columns=x_columns, num_points=num_points, num_threads=num_threads, echo=echo
		)
